"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("@dcfjs/common/noCaptureEnv");
const RDD_1 = require("./RDD");
const client_1 = require("@dcfjs/common/client");
const http2 = require("http2");
const sf = require("@dcfjs/common/serializeFunction");
const ProgressBar = require("progress");
const split = require("split");
const v8 = require('v8');
const defaultOption = {
    showProgress: !!process.stdout.isTTY,
    defaultPartitions: 4,
    defaultPersistStorage: 'disk',
    defaultRepartitionStorage: 'disk',
};
class Context {
    constructor(client, option) {
        this._fileLoaders = [];
        this._client = client;
        this._option = {
            ...defaultOption,
            ...option,
        };
        this._client.session.on('stream', (stream, header) => {
            const path = header[http2.constants.HTTP2_HEADER_PATH];
            switch (path) {
                case '/progress': {
                    let progress;
                    stream
                        .pipe(split())
                        .on('data', function (line) {
                        if (!line) {
                            return;
                        }
                        const [curr, total] = v8.deserialize(Buffer.from(line, 'base64'));
                        if (!progress) {
                            progress = new ProgressBar(`:percent [:bar] Partitions :current/:total :rate/s :etas`, {
                                clear: true,
                                curr,
                                total,
                            });
                        }
                        else {
                            progress.tick();
                        }
                    })
                        .on('end', () => {
                        if (progress) {
                            progress.terminate();
                            progress = null;
                        }
                    });
                    break;
                }
                default: {
                    stream.close();
                }
            }
        });
    }
    get client() {
        return this._client;
    }
    get option() {
        return this._option;
    }
    registerFileLoader(loader) {
        this._fileLoaders.push(loader);
    }
    close() {
        return this._client.close();
    }
    async execute(numPartitions, partitionFunc, // first function run on worker, second function run on master
    finalFunc) {
        const { showProgress } = this._option;
        const ret = await this._client.post('/exec', sf.serializeFunction(sf.captureEnv((async (dispatchWork, createPushStream, tempStorageSession) => {
            let sendProgress;
            let stream;
            if (showProgress) {
                stream = await createPushStream('/progress');
                sendProgress = (current, total) => {
                    stream.write(v8.serialize([current, total]).toString('base64') + '\n');
                };
            }
            else {
                sendProgress = () => { };
            }
            let current = 0;
            sendProgress(0, numPartitions);
            function tickAfter(p) {
                p.then(() => {
                    sendProgress(++current, numPartitions);
                });
                return p;
            }
            try {
                const partitionResults = await Promise.all(new Array(numPartitions).fill(0).map((v, i) => {
                    const f = partitionFunc(i, tempStorageSession);
                    if (Array.isArray(f)) {
                        return tickAfter(dispatchWork(f[0]).then(f[1]));
                    }
                    return tickAfter(dispatchWork(f));
                }));
                return await finalFunc(partitionResults, tempStorageSession);
            }
            finally {
                if (stream) {
                    await new Promise(resolve => stream.end(resolve));
                }
            }
        }), {
            numPartitions,
            partitionFunc,
            finalFunc,
            showProgress,
            v8: sf.requireModule('v8'),
        })));
        return ret;
    }
    emptyRDD() {
        return new RDD_1.GeneratedRDD(this, 0);
    }
    range(from, to, step = 1, numPartitions) {
        if (to == null) {
            to = from;
            from = 0;
        }
        numPartitions = numPartitions || this._option.defaultPartitions;
        const finalCount = Math.ceil((to - from) / step);
        const rest = finalCount % numPartitions;
        const eachCount = (finalCount - rest) / numPartitions;
        const args = [];
        let index = 0;
        for (let i = 0; i < numPartitions; i++) {
            const subCount = i < rest ? eachCount + 1 : eachCount;
            const end = index + subCount;
            args.push({
                from: from + step * index,
                count: subCount,
            });
            index = end;
        }
        return new RDD_1.GeneratedRDD(this, numPartitions, sf.captureEnv(partitionId => {
            const { from, count } = args[partitionId];
            return sf.captureEnv(() => {
                const ret = [];
                for (let i = 0; i < count; i++) {
                    ret.push(from + step * i);
                }
                return ret;
            }, { from, count, step });
        }, { step, args, sf: sf.requireModule('@dcfjs/common/serializeFunction') }));
    }
    parallelize(arr, numPartitions) {
        numPartitions = numPartitions || this._option.defaultPartitions;
        const args = [];
        const rest = arr.length % numPartitions;
        const eachCount = (arr.length - rest) / numPartitions;
        let index = 0;
        for (let i = 0; i < numPartitions; i++) {
            const subCount = i < rest ? eachCount + 1 : eachCount;
            const end = index + subCount;
            args.push(arr.slice(index, end));
            index = end;
        }
        return new RDD_1.GeneratedRDD(this, numPartitions, sf.captureEnv(partitionId => {
            const data = args[partitionId];
            return sf.captureEnv(() => data, {
                data,
            });
        }, {
            args,
            sf: sf.requireModule('@dcfjs/common/serializeFunction'),
        }));
    }
    union(...rdds) {
        return new RDD_1.UnionRDD(this, rdds);
    }
    getFileLoader(baseUri) {
        return this._fileLoaders.find(v => v.canHandleUri(baseUri));
    }
    binaryFiles(baseUri, { recursive = false, } = {}) {
        const loader = this.getFileLoader(baseUri);
        if (!loader) {
            throw new Error('No loader can handle uri: `' + baseUri + '`');
        }
        const { listFiles, loadFile } = loader;
        return new RDD_1.FileLoaderRDD(this, sf.captureEnv(() => {
            return listFiles(baseUri, recursive);
        }, {
            listFiles,
            baseUri,
            recursive,
        }), sf.captureEnv((fileName) => {
            return loadFile(baseUri, fileName);
        }, {
            loadFile,
            baseUri,
        }));
    }
    wholeTextFiles(baseUrl, { decompressor, encoding = 'utf8', recursive = false, functionEnv, } = {}) {
        if (typeof encoding === 'boolean') {
            recursive = encoding;
            encoding = 'utf-8';
        }
        if (typeof decompressor === 'function') {
            sf.captureEnv(decompressor, functionEnv);
        }
        return this.binaryFiles(baseUrl, { recursive }).mapPartitions(sf.captureEnv(v => {
            let buf = v[0][1];
            if (decompressor) {
                buf = decompressor(buf, v[0][0]);
            }
            return [[v[0][0], buf.toString(encoding)]];
        }, { encoding, decompressor }));
    }
    textFile(baseUrl, options) {
        const { __dangerousDontCopy: dontCopy = false } = options || {};
        return this.wholeTextFiles(baseUrl, options).flatMap(sf.captureEnv((v) => {
            const ret = v[1].replace(/\\r/m, '').split('\n');
            // Remove last empty line.
            if (!ret[ret.length - 1]) {
                ret.pop();
            }
            if (dontCopy) {
                return ret;
            }
            // Fix memory leak: sliced string keep reference of huge string
            // see https://bugs.chromium.org/p/v8/issues/detail?id=2869
            return ret.map((v) => (' ' + v).substr(1));
        }, {
            dontCopy,
        }));
    }
}
exports.Context = Context;
async function createContext(masterEndpoint) {
    const context = new Context(await client_1.createClient(masterEndpoint));
    return context;
}
exports.createContext = createContext;
